from os.path import join as pathjoin, dirname, abspath


import cv2
import pytest
import py.path


def test_image_abs_path(image_path):
    full_path = py.path.local(abspath(__file__)).dirpath("data/images", image_path)
    if not full_path.check(file=1):
        pytest.skip("Test data file not found. Retrieve docoskin-test-data to run full test suite.")
    return str(full_path)


def read_test_image(image_path):
    return cv2.imread(test_image_abs_path(image_path), cv2.IMREAD_GRAYSCALE)


def open_test_image(image_path):
    return open(test_image_abs_path(image_path), "r")



from concurrent.futures import ThreadPoolExecutor
import logging

import pytest


@pytest.yield_fixture(scope="session", params=(None, 2,))
def thread_pool(request):
    thread_pool = ThreadPoolExecutor(request.param) if request.param is not None else None

    yield thread_pool

    if thread_pool is not None:
        thread_pool.shutdown()


@pytest.fixture(scope="session", autouse=True)
def debug_logging(request):
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s:%(levelname)s:%(name)s:%(threadName)s:%(message)s",
    )



"""
    Contains apparatus to allow docoskin to run in an unthreaded mode (without knowing about it)
"""
from __future__ import absolute_import

from six.moves import map as imap


class DummyFuture(object):
    def __init__(self, result, exception=None):
        self._result = result
        self._exception = exception

    def cancel(self):
        return False

    def cancelled(self):
        return False

    def done(self):
        return True

    def result(self, timeout=None):
        if self._exception is not None:
            raise self._exception
        return self._result

    def exception(self, timeout=None):
        return self._exception

    def add_done_callback(self, fn):
        # emulate "future has already returned" behaviour
        fn()


class DummyThreadPoolExecutor(object):
    def submit(self, fn, *args, **kwargs):
        exception = result = None
        try:
            result = fn(*args, **kwargs)
        except Exception as e:
            exception = e

        return DummyFuture(result, exception=exception)

    def map(self, func, *iterables, **kwargs):
        return imap(func, *iterables)

    def shutdown(self, wait=True):
        pass



from __future__ import absolute_import

from concurrent.futures import ThreadPoolExecutor
import logging

import cv2
import numpy

from docoskin import defaults
from docoskin.image_operations import coverage_from_candidate_warp, diff_overlay_images, stretched_contrast
from docoskin.warping import match_and_warp_candidate


logger = logging.getLogger("docoskin")


_feature_matcher_factories = {
    "bruteforce": defaults.default_bruteforce_feature_matcher_factory,
    "flann": defaults.default_flann_feature_matcher_factory,
}


def docoskin(
        reference_image,
        candidate_image,
        out_file,
        contrast_stretch=True,
        warped_candidate_out_file=None,
        feature_matcher=None,
        thread_pool=None,
        ):
    # we use a combination of numpy and imdecode/imencode for file handling as it allows us to transparently work with
    # any file-like object (including stdin and stdout through "-" options to argparse)
    reference_image = original_reference_image = cv2.imdecode(
        numpy.fromfile(reference_image, dtype="uint8"),
        cv2.IMREAD_GRAYSCALE,
    )
    candidate_image = original_candidate_image = cv2.imdecode(
        numpy.fromfile(candidate_image, dtype="uint8"),
        cv2.IMREAD_GRAYSCALE,
    )

    if contrast_stretch:
        logger.debug("Stretching contrast for reference_image")
        reference_image = stretched_contrast(reference_image)
        logger.debug("Stretching contrast for candidate_image")
        candidate_image = stretched_contrast(candidate_image)

    warped_candidate, M = match_and_warp_candidate(
        reference_image,
        candidate_image,
        warp_image=original_candidate_image,
        feature_matcher=feature_matcher,
        thread_pool=thread_pool,
    )

    if contrast_stretch:
        # the darkest or lightest regions of the candidate image may now have been transformed off the image area so
        # we re-apply the contrast stretching to get results calculated based just on the page area
        coverage = coverage_from_candidate_warp(reference_image.shape, candidate_image.shape, M)
        logger.debug("Warped candidate coverage = %s", coverage)
        logger.debug("Re-stretching contrast for warped candidate")
        warped_candidate = stretched_contrast(warped_candidate, coverage=coverage)
    if warped_candidate_out_file:
        cv2.imencode(".png", warped_candidate)[1].tofile(warped_candidate_out_file)

    overlayed_candidate = diff_overlay_images(reference_image, warped_candidate)

    cv2.imencode(".png", overlayed_candidate)[1].tofile(out_file)


def main():
    import argparse

    def nonneg_int(value):
        n = int(value)
        if n < 0:
            raise ValueError("Value cannot be negative")
        return n

    parser = argparse.ArgumentParser(
        description="Onion-skin two document images and output resulting png file to stdout"
    )
    parser.add_argument("reference_image", type=argparse.FileType("r"), help="Reference document image file")
    parser.add_argument("candidate_image", type=argparse.FileType("r"), help="Candidate document image file")
    parser.add_argument(
        "out_image",
        type=argparse.FileType("w"),
        nargs="?",
        default="-",
        help="Candidate document image file (default: %(default)s)",
    )
    parser.add_argument(
        "--matcher", "-m",
        choices=_feature_matcher_factories.keys(),
        default="flann",
        help="Select which feature matcher to use",
    )
    parser.add_argument(
        "--threads", "-t",
        type=nonneg_int,
        default=2,
        metavar="N",
        help="Number of python threads to use (default: %(default)s). 0 disables threading, opencv may use more " \
            "threads internally",
    )
    parser.add_argument(
        "--warped-candidate-out", "-w",
        type=argparse.FileType("w"),
        metavar="FILE",
        help="Output warped (but unmerged) candidate image to FILE",
    )
    parser.add_argument(
        "--no-contrast-stretch", "-C",
        dest="contrast_stretch",
        action="store_false",
        help="Don't perform any contrast-stretching steps on input files",
    )
    parser.add_argument("--verbose", "-v", action="store_true", help="Emit debugging information")
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    docoskin(
        args.reference_image,
        args.candidate_image,
        args.out_image,
        contrast_stretch=args.contrast_stretch,
        warped_candidate_out_file=args.warped_candidate_out,
        feature_matcher=_feature_matcher_factories[args.matcher](),
        thread_pool=ThreadPoolExecutor(args.threads) if args.threads else None,
    )


if __name__ == "__main__":
    main()



from __future__ import absolute_import

from collections import namedtuple
from itertools import compress
import logging

import cv2
import numpy

from docoskin import defaults
from docoskin.dummythreadpoolexecutor import DummyThreadPoolExecutor
from docoskin.exceptions import DocoskinInvalidArgumentCombinationError, DocoskinNoMatchFoundError


logger = logging.getLogger("docoskin")


FindCandidateHomographyExtendedOutput = namedtuple(
    "FindCandidateHomographyExtendedOutput",
    (
        "reference_keypoints",
        "candidate_keypoints",
        "inlier_matches",
        "outlier_matches",
        "bad_matches",
    ),
)


_KeypointsDescriptorsTuple = namedtuple("KeypointsDescriptorsTuple", ("keypoints", "descriptors"))


def _detect_and_compute(feature_detector_factory, image, name=None):
    r = _KeypointsDescriptorsTuple(*feature_detector_factory().detectAndCompute(image, None))
    logger.debug("Found %i keypoints%s%s", len(r.keypoints), " in " if name else "", name)
    return r


def find_candidate_homography(
        reference_image,
        candidate_image,
        feature_detector_factory=None,
        feature_matcher=None,
        feature_distance_ratio_threshold=defaults.DEFAULT_FEATURE_DISTANCE_RATIO_THRESHOLD,
        n_match_threshold=defaults.DEFAULT_N_MATCH_THRESHOLD,
        reference_keypoints=None,
        reference_descriptors=None,
        ransac_reproj_threshold=defaults.DEFAULT_RANSAC_REPROJ_THRESHOLD,
        ransac_n_inlier_threshold=defaults.DEFAULT_RANSAC_N_INLIER_THRESHOLD,
        thread_pool=None,
        ):
    feature_detector_factory = feature_detector_factory or defaults.default_feature_detector_factory
    feature_matcher = feature_matcher or defaults.default_feature_matcher_factory()
    thread_pool = thread_pool or DummyThreadPoolExecutor()

    candidate_kp_dsc_f = thread_pool.submit(
        _detect_and_compute,
        feature_detector_factory,
        candidate_image,
        "candidate",
    )

    already_trained_descriptors = feature_matcher.getTrainDescriptors()
    if already_trained_descriptors:
        if len(reference_keypoints or ()) != len(already_trained_descriptors):
            raise DocoskinInvalidArgumentCombinationError(
                "Pre-trained feature matchers require a reference_keypoints argument containing the corresponding "
                "keypoints"
            )
    else:
        if not (reference_keypoints or reference_descriptors):
            reference_kp_dsc_f = thread_pool.submit(
                _detect_and_compute,
                feature_detector_factory,
                reference_image,
                "reference",
            )
            # no point in waiting the reference_kp_dsc_f, we need it for the next step
            reference_keypoints, reference_descriptors = reference_kp_dsc_f.result()
        elif reference_keypoints and reference_descriptors:
            if len(reference_keypoints) != len(reference_descriptors):
                raise DocoskinInvalidArgumentCombinationError(
                    "reference_keypoints and reference_descriptors length mismatch"
                )
        else:
            raise DocoskinInvalidArgumentCombinationError(
                "Doesn't make sense to supply reference_keypoints without reference_descriptors or vice-versa"
            )

        feature_matcher.add((reference_descriptors,))
        # so now reference_keypoints and reference_descriptors should have been defined and feature_matcher should have
        # its reference_descriptors one way or another

    matches = feature_matcher.knnMatch(candidate_kp_dsc_f.result().descriptors, k=2)

    # candidate_kp_dsc_f must have returned by now to get to this point, so let's give them some more conventient
    # accessors
    candidate_keypoints, candidate_descriptors = candidate_kp_dsc_f.result()

    good_match_mask = tuple(
        match and (len(match) == 1 or match[0].distance < feature_distance_ratio_threshold*match[1].distance)
        for match in matches
    )
    n_good_matches = sum(1 for good in good_match_mask if good)
    logger.debug("Found %i/%i 'good' keypoint matches", n_good_matches, len(matches))
    if n_good_matches < n_match_threshold:
        raise DocoskinNoMatchFoundError("Not enough 'good' feature matches found ({})".format(n_good_matches))

    reference_coords = numpy.float32(tuple(
        reference_keypoints[match[0].trainIdx].pt for match in compress(matches, good_match_mask)
    )).reshape(-1, 1, 2,)
    candidate_coords = numpy.float32(tuple(
        candidate_keypoints[match[0].queryIdx].pt for match in compress(matches, good_match_mask)
    )).reshape(-1, 1, 2,)

    M, inlier_mask = cv2.findHomography(candidate_coords, reference_coords, cv2.RANSAC, ransac_reproj_threshold)
    n_inliers = sum(inlier_mask.flat)
    if n_inliers < ransac_n_inlier_threshold:
        raise DocoskinNoMatchFoundError("Not enough RANSAC inliers found ({})".format(n_inliers))

    logger.debug("Used %i keypoints as inliers", n_inliers)

    extended_output = FindCandidateHomographyExtendedOutput(
        reference_keypoints=reference_keypoints,
        candidate_keypoints=candidate_keypoints,
        inlier_matches=tuple(
            match[0] for match in compress(compress(matches, good_match_mask), inlier_mask.flat)
        ),
        outlier_matches=tuple(
            match[0] for match in compress(compress(matches, good_match_mask), ((not inlier) for inlier in inlier_mask.flat))
        ),
        bad_matches=tuple(
            match[0] for match in compress(matches, ((not good) for good in good_match_mask))
        ),
    )

    return M, extended_output


def match_and_warp_candidate(reference_image, candidate_image, warp_image=None, **kwargs):
    M = find_candidate_homography(reference_image, candidate_image, **kwargs)[0]
    return cv2.warpPerspective(
        warp_image if warp_image is not None else candidate_image,
        M,
        tuple(reversed(reference_image.shape)),
        flags=cv2.INTER_CUBIC,
        borderMode=cv2.BORDER_CONSTANT,
        borderValue=0,
    ), M



from itertools import product
import logging
from math import hypot


import cv2
from docoskin.warping import find_candidate_homography
import numpy
import pytest
from six.moves import zip as izip


from utils import read_test_image


@pytest.mark.parametrize("reference_path,candidate_path,ref_corners_in_cand,tolerance", (
    ("words-to-avoid/reference.png", "words-to-avoid/candidate01.png", (
        (1557, 180), (1426, 895),
        (408, 41), (370, 751),
    ), 5,), # didn't keep the original of this image - hard to determine the "ground truth" corner coords. up tolerance.
    ("words-to-avoid/reference.png", "words-to-avoid/candidate02.png", (
        (2004, 2044), (811, 2044),
        (2004, 484), (811, 484),
    ), 5,),
    ("words-to-avoid/reference.png", "words-to-avoid/candidate03.png", (
        (1793, 2573), (6, 2523),
        (1820, 26), (45, 15),
    ), 5,),
    ("words-to-avoid/reference.png", "words-to-avoid/candidate04.png", (
        (254, 490), (1672, 457),
        (319, 2963), (1925, 2852),
    ), 5,),
    ("words-to-avoid/reference.png", "words-to-avoid/candidate05.png", (
        (2832, 22), (3140, 1368),
        (469, 604), (904, 1930),
    ), 5,),
    ("why-ten-years/reference.png", "why-ten-years/candidate01.png", (
        (-25, -25), (790, 14),
        (13, 1228), (800, 1242),
    ), 5,),
    ("why-ten-years/reference.png", "why-ten-years/candidate02.png", (
        (1301, 998), (627, 991),
        (1242, 11), (605, 27),
    ), 5,),
    ("why-ten-years/reference.png", "why-ten-years/candidate03.png", (
        (629, 285), (1522, 545),
        (272, 1771), (1298, 1970),
    ), 5,),
))
def test_matrix_generation(reference_path, candidate_path, ref_corners_in_cand, tolerance, thread_pool,):
    reference_image = read_test_image(reference_path)
    candidate_image = read_test_image(candidate_path)

    M = find_candidate_homography(reference_image, candidate_image, thread_pool=thread_pool)[0]

    # opencv wants its coords in a slightly odd format for this call
    ref_corners_in_cand_a = numpy.float32((ref_corners_in_cand,))

    # ref_corners_in_cand is supposed to denote the 4 corners of the reference document in the candidate image
    # specified in the order top left, top right, bottom left, bottom right. so if all is well those coordinates passed
    # through our recovered perspective matrix should be fairly close to reference_image's dimensions
    transformed_coords = cv2.perspectiveTransform(ref_corners_in_cand_a, M)[0]
    assert all(
        # note ref_y, ref_x order switcharoo here because product() would otherwise traverse the corners in the opposite
        # order we want
        hypot(tr_x-ref_x, tr_y-ref_y) < tolerance for (tr_x, tr_y), (ref_y, ref_x) in izip(
            transformed_coords,
            product((0, reference_image.shape[0],), (0, reference_image.shape[1],)),
        )
    )



from __future__ import absolute_import

import cv2


DEFAULT_AKAZE_THRESHOLD=0.005
DEFAULT_FEATURE_DISTANCE_RATIO_THRESHOLD=0.7
DEFAULT_N_MATCH_THRESHOLD=10
DEFAULT_RANSAC_REPROJ_THRESHOLD=5.0
DEFAULT_RANSAC_N_INLIER_THRESHOLD=4
DEFAULT_REMOVED_COLOR_MATRIX=((1.0, 1.0, 0.0,), (0.0, 0.0, 1.0,),)
DEFAULT_ADDED_COLOR_MATRIX=((0.0, 1.0, 0.0,), (1.0, 0.0, 1.0,),)
DEFAULT_CONTRAST_STRETCH_LOWER_PERCENTILE=3.0
DEFAULT_CONTRAST_STRETCH_UPPER_PERCENTILE=92.5


def default_feature_detector_factory():
    return cv2.AKAZE_create(threshold=DEFAULT_AKAZE_THRESHOLD)


def default_bruteforce_feature_matcher_factory():
    return cv2.BFMatcher(normType=cv2.NORM_HAMMING)


def default_flann_feature_matcher_factory():
    return cv2.FlannBasedMatcher(
        indexParams={
            "algorithm": 6,  # FLANN_INDEX_LSH
            "table_number": 6,
            "key_size": 14,
            "multi_probe_level": 1,
        },
        searchParams={
            "checks": 32,
        },
    )


default_feature_matcher_factory = default_flann_feature_matcher_factory



import logging
import os


import cv2
from docoskin import docoskin
import numpy
import pytest


from utils import open_test_image, read_test_image


_words_to_avoid_bboxes = (
    ((132, 2604), (146, 2618),),
    ((159, 2634), (173, 2648),),
    ((159, 2604), (173, 2618),),
    ((132, 2634), (146, 2648),),
)
_why_ten_years_bboxes = (
    ((755, 24), (768, 37),),
    ((782, 54), (795, 67),),
    ((782, 24), (795, 37),),
    ((755, 54), (768, 67),),
)


@pytest.mark.parametrize("reference_path,candidate_path,black_bbox,white_bbox,added_bbox,removed_bbox", (
    (
        "words-to-avoid/reference.png",
        "words-to-avoid/candidate01.png",
    ) + _words_to_avoid_bboxes,
    (
        "words-to-avoid/reference.png",
        "words-to-avoid/candidate02.png",
    ) + _words_to_avoid_bboxes,
    (
        "words-to-avoid/reference.png",
        "words-to-avoid/candidate03.png",
    ) + _words_to_avoid_bboxes,
    (
        "words-to-avoid/reference.png",
        "words-to-avoid/candidate04.png",
    ) + _words_to_avoid_bboxes,
    (
        "words-to-avoid/reference.png",
        "words-to-avoid/candidate05.png",
    ) + _words_to_avoid_bboxes,
    (
        "why-ten-years/reference.png",
        "why-ten-years/candidate01.png",
    ) + _why_ten_years_bboxes,
    (
        "why-ten-years/reference.png",
        "why-ten-years/candidate02.png",
    ) + _why_ten_years_bboxes,
    (
        "why-ten-years/reference.png",
        "why-ten-years/candidate03.png",
    ) + _why_ten_years_bboxes,
))
def test_full_pipeline(
        tmpdir,
        reference_path,
        candidate_path,
        black_bbox,
        white_bbox,
        added_bbox,
        removed_bbox,
        thread_pool,
        ):
    reference_image = read_test_image(reference_path)
    reference_image_shape = reference_image.shape
    del reference_image  # allow python to recover this memory

    reference_image_file = open_test_image(reference_path)
    candidate_image_file = open_test_image(candidate_path)

    out_file = tmpdir.join("out.png").ensure(file=True).open("r+")

    docoskin(reference_image_file, candidate_image_file, out_file, thread_pool=thread_pool)

    out_file.seek(0)

    recovered_image = cv2.imdecode(numpy.fromfile(out_file, dtype="uint8"), cv2.IMREAD_COLOR)
    assert recovered_image.shape[:2] == reference_image_shape[:2]

    for ((x0, y0), (x1, y1)), target_bgr in (
            (black_bbox, (0, 0, 0,)),
            (white_bbox, (255, 255, 255,)),
            (added_bbox, (0, 255, 0,)),
            (removed_bbox, (0, 0, 255,)),
        ):
        region_pixels = recovered_image[y0:y1,x0:x1].reshape((-1,3))
        pixel_value_deltas = numpy.float32(target_bgr) - region_pixels
        pixel_value_distances = numpy.sqrt(numpy.sum(pixel_value_deltas**2, axis=1))
        assert numpy.percentile(pixel_value_distances, 75.0, interpolation="lower") < 16



from itertools import product
import logging
from math import floor


import cv2
from docoskin.image_operations import stretched_contrast
import numpy
import pytest


@pytest.mark.parametrize("lower_percentile,upper_percentile,uncovered_rows,shuffle_pixels", tuple(product(
    (0, 1, 5, 14.3, 45,),
    (77, 90, 92.7, 99, 100,),
    (0, 13, 121,),
    (False, True,),
)))
def test_stretched_contrast(lower_percentile, upper_percentile, uncovered_rows, shuffle_pixels):
    image = numpy.broadcast_to(numpy.arange(256, dtype="uint8"), (256, 256))

    if shuffle_pixels:
        # this should make no difference to the assertion phase, so no additional changes needed beyond this
        numpy.random.seed(1234)
        flattened_image = image.flatten()
        numpy.random.shuffle(flattened_image)
        image = flattened_image.reshape(image.shape)

    if uncovered_rows:
        # place `uncovered_rows` black rows at the bottom of the image
        padded_image = numpy.concatenate((image, numpy.zeros((uncovered_rows, 256,), dtype="uint8"),), axis=0)
    else:
        padded_image = image

    adjusted_image = stretched_contrast(
        padded_image,
        lower_percentile=lower_percentile,
        upper_percentile=upper_percentile,
        coverage=1.0 if not uncovered_rows else (256.0/(256+uncovered_rows)),
    )

    if uncovered_rows:
        # strip the uncovered rows back off for analysis
        adjusted_image = adjusted_image[:256,:]

    assert ((adjusted_image == 0) == (image <= floor((float(lower_percentile)/100)*256))).all()
    # the following comparison value is clipped to 255 (via min()) as we are expected to accept 100.0 percent as a
    # maximum makes the input a closed interval, but we of course are working against half-open interval data.
    assert ((adjusted_image == 255) == (image >= min(floor((float(upper_percentile)/100)*256), 255))).all()

    # the 100's cancel out in this expected_new_median calculation
    expected_new_median =  256 * (50.0-lower_percentile)/(upper_percentile-lower_percentile)
    # we should allow a tolerance for quantization difficulties, but this should vary with the amount of "gain" that's
    # expected to be applied to the image
    tolerance = 1.5 * 100.0/(upper_percentile-lower_percentile)
    assert abs(numpy.percentile(adjusted_image,  50.0, interpolation="lower") - expected_new_median) <= tolerance



from __future__ import absolute_import

from itertools import product

import logging

import cv2
import numpy

from docoskin import defaults


logger = logging.getLogger("docoskin")


def coverage_from_candidate_warp(reference_image_shape, candidate_image_shape, M):
    # first check if given this projection M, any corners of reference_image lie outside the extents of candidate_image
    if all(
            0 <= x < candidate_image_shape[1] and 0 <= y < candidate_image_shape[0]
            for x, y in cv2.perspectiveTransform(
                # did i mention perspectiveTransform expects a weird coordinate array format?
                numpy.float32((tuple(product((0, reference_image_shape[1],), (0, reference_image_shape[0],))),)),
                numpy.linalg.inv(M),
            )[0]
            ):
        # they don't - let's not waste our time calculating anything else
        return 1.0

    # "count" the number of pixels that would make it onto the reference_image canvas
    return cv2.warpPerspective(
        numpy.ones(candidate_image_shape),
        M,
        tuple(reversed(reference_image_shape)),
        flags=cv2.INTER_NEAREST,
        borderMode=cv2.BORDER_CONSTANT,
        borderValue=0,
    ).sum(dtype="float32") / (reference_image_shape[0] * reference_image_shape[1])


def stretched_contrast(
        image,
        lower_percentile=defaults.DEFAULT_CONTRAST_STRETCH_LOWER_PERCENTILE,
        upper_percentile=defaults.DEFAULT_CONTRAST_STRETCH_UPPER_PERCENTILE,
        coverage=1.0,  # the proportion of the image which is covered by relevant pixels (non-relevant pixels are
                       # expected to be value 0)
        ):
    if lower_percentile is None:
        black_point = 0
    else:
        black_point = numpy.percentile(
            image,
            (100 * (1.0-coverage)) + (lower_percentile*coverage),
            interpolation="lower",
        )
    logger.debug("Black point for image at percentile %r = %s", lower_percentile, black_point)

    if upper_percentile is None:
        white_point = 255
    else:
        white_point = numpy.percentile(
            image,
            (100 * (1.0-coverage)) + (upper_percentile*coverage),
            # "lower" is arguably the right choice here, but the reality is that "lower" is the easier behaviour to
            # predict/test against
            interpolation="lower",
        )
    logger.debug("White point for image at percentile %r = %s", upper_percentile, white_point)

    out_image = numpy.clip(
        # careful to convert to float *before* subtraction otherwise we may get wrap-around from any sub-black-point
        # pixels that become negative if the calculation is done in uint8
        (image.astype("float32")-black_point)*(256.0/(white_point-black_point)),
        0,
        255,
    ).astype("uint8")
    return out_image


def diff_overlay_images(
        reference_image,
        candidate_image,
        # a reminder here that opencv uses BGR order for some reason and this reflected in the order these matrices are
        # specified
        removed_color_matrix=defaults.DEFAULT_REMOVED_COLOR_MATRIX,
        added_color_matrix=defaults.DEFAULT_ADDED_COLOR_MATRIX,
        ):
    # combine images into a single two-channel image
    stacked_image = numpy.stack((reference_image, candidate_image,), axis=-1)

    # now we can apply a matrix multiplication to produce two RGB images, highlighting the removed and added sections
    # respectively. we are essentially performing a linear transformation of the planar, 2d colorspace into the 3d
    # RGB cube if it help to visualize it that way.
    removed_image = numpy.dot(stacked_image, numpy.float32(removed_color_matrix)).astype("uint8")
    added_image = numpy.dot(stacked_image, numpy.float32(added_color_matrix)).astype("uint8")

    # the final image is produced by selecting between these two RGB images depending on whether each pixel underwent a
    # removal (less black) or an addition (more black). we determine this using a simple pixel value comparison
    return numpy.where(numpy.expand_dims(stacked_image[:,:,0] > stacked_image[:,:,1], -1), added_image, removed_image)



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))


with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


version_namespace = {}
with open(path.join(here, "docoskin/version.py")) as f:
    exec(f.read(), version_namespace)


setup(
    name="docoskin",
    version=version_namespace["__version__"],

    description='"Onion-skin" visual differences between a reference document image and a scanned copy',
    long_description=long_description,
    long_description_content_type="text/markdown",

    url='https://github.com/risicle/docoskin',

    author="Robert Scott",
    author_email="code@humanleg.org.uk",

    license="GPLv3",

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Legal Industry',
        'Topic :: Multimedia :: Graphics :: Viewers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    keywords="document scan compare visual diff signature",
    packages=["docoskin"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    install_requires=[
        # omitting opencv >= 3.1.0 dependency as there is no standard python package for it and therefore no universal
        # way to detect it. i advocate using nix to satisfy all these dependencies, hence the included default.nix.
        "six",
    ],
    extras_require={
        ':python_version<"3"': ["futures"],
        "pypi_opencv": ["opencv-python >= 3.1.0"],
    },

    entry_points={
        'console_scripts': [
            'docoskin=docoskin.docoskin:main',
        ],
    },
)
